package v2

import (
	"context"
	"fmt"

	uuid "github.com/satori/go.uuid"
	"k8s.io/apimachinery/pkg/api/errors"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	quarkv1alpha1 "quark.netapp.io/volume-controller/pkg/apis/volumecontroller/v1alpha1"
	"quark.netapp.io/volume-controller/pkg/log"
	"quark.netapp.io/volume-controller/pkg/quarkconfig"
)

// nvrBackup implements the nvrInterface to support the backup usecase. Functionalities like break/resync which are not
// planned for the first release of Integrated backup feature should return error until implemented.
type nvrBackup struct {
	nvrCommonImpl
}

func (r *nvrBackup) Create() (*quarkv1alpha1.NetAppVolumeReplication, error) {
	storagePoolUUIDStr := r.nvr.Labels[quarkconfig.StoragePoolLabel]
	_, err := r.nvcClient.NetAppStoragePoolGet(storagePoolUUIDStr)
	if err != nil {
		return nil, fmt.Errorf("error getting StoragePool %v: %v", storagePoolUUIDStr, err)
	}

	objectStoreUUIDStr := r.nvr.Labels[quarkconfig.ObjectStoreLabel]
	// not validating err in the conversion below since if uuid string is part of labels in the cr, it means it has already been vetted earlier
	objectStoreUUID, _ := uuid.FromString(objectStoreUUIDStr)
	_, err = r.nvcClient.NetAppObjectStoreGet(objectStoreUUID)
	if err != nil {
		return nil, fmt.Errorf("error getting ObjectStore %v: %v", objectStoreUUIDStr, err)
	}

	// "Mirrored" is the only possible state for backup.
	if r.nvr.Spec.MirrorState != quarkv1alpha1.MirrorStateMirrored {
		return nil, fmt.Errorf("invalid mirror state \"%v\", only allowed state is \"%v\"", r.nvr.Spec.MirrorState, quarkv1alpha1.MirrorStateMirrored)
	}

	if r.nvr.Spec.Source == nil {
		return nil, fmt.Errorf("Source field in spec is nil, cannot process create request")
	}
	srcVolUUIDStr := r.nvr.Spec.Source.UUID
	srcVol, err := r.nvcClient.NetAppVolumeGet(srcVolUUIDStr)
	if err != nil {
		return nil, fmt.Errorf("error getting source volume %v: %v", srcVolUUIDStr, err)
	}

	// Check if destination volume belongs to the storagepool supplied in the replication cr
	if srcVol.Labels[quarkconfig.StoragePoolLabel] != storagePoolUUIDStr {
		return nil, fmt.Errorf("source volume StoragePool (%v) not the same as StoragePool supplied in custom resource (%v)", srcVol.Labels[quarkconfig.StoragePoolLabel], storagePoolUUIDStr)
	}

	ntapVolSnapList, err := r.nvcClient.NetAppVolumeSnapshotsListGet(srcVol.Name)
	if err != nil {
		return nil, fmt.Errorf("error getting snapshots for volume %v", srcVol.Name)
	}

	foundSnapshotToTransfer := false
	for _, snap := range ntapVolSnapList.Status.Snapshots {
		if snap.Name == r.nvr.Spec.SnapshotToTransfer {
			// if snapshot is in snapList, it is ready and therefore can be backed up
			foundSnapshotToTransfer = true
			break
		}
	}

	if !foundSnapshotToTransfer {
		return nil, fmt.Errorf("requested snapshot %v does not exist for volume %v", r.nvr.Spec.SnapshotToTransfer, srcVol.Name)
	}

	ownerID := r.nvr.Labels[quarkconfig.OwnerIDLabel]
	storagePoolUUID, _ := uuid.FromString(storagePoolUUIDStr)
	volReplUUID, _ := uuid.FromString(r.nvr.Name)
	srcVolUUID, _ := uuid.FromString(srcVolUUIDStr)
	destUUID, _ := uuid.FromString(r.nvr.Spec.Destination.UUID)
	cloudbackupList, err := constructNetAppVolumeCloudBackupList(
		ownerID,
		destUUID,
		storagePoolUUID,
		srcVolUUID,
		objectStoreUUID,
		volReplUUID,
	)
	if err != nil {
		return nil, err
	}

	nvr, err := r.nvcClient.quarkClient.QuarkV1alpha1().NetAppVolumeReplications(r.nvcClient.nvcNamespace).Create(
		context.TODO(), r.nvr, metav1.CreateOptions{})
	if err != nil {
		// If already created in a previous attempt, don't error out and continue with operations.
		if !errors.IsAlreadyExists(err) {
			return nil, err
		}
		log.WithFields(log.Fields{
			"sourceVolume":      r.nvr.Spec.Source.UUID,
			"volumeReplication": r.nvr.Name,
		}).Debug("NetAppVolumeReplication (backup) CR already exists")
	}

	// cloudbackupList will have owner ref set to the nvr created above
	cloudbackupList, err = r.nvcClient.netAppVolumeCloudBackupListCreate(cloudbackupList)
	if err != nil {
		// If already exists, return without error, idempotent behavior of this API.
		if errors.IsAlreadyExists(err) {
			log.WithFields(log.Fields{
				"sourceVolume":      r.nvr.Spec.Source.UUID,
				"volumeReplication": r.nvr.Name,
			}).Debug("NetAppVolumeCloudBackupList for volume already exists")
			return nvr, nil
		}
		log.Errorf("failed to create CloudBackupList %v, attempting to delete VolumeReplication %v which was already created",
			cloudbackupList.Name, nvr.Name)
		// try and delete nvr created above but not the end of world if it fails. Note that its going to be
		// very rare that api server went down immediately after successfully creating nvr in the previous step. In case
		// it went down, the delete of nvr will fail as well but as a good citizen we can try.
		delNVRErr := r.nvcClient.quarkClient.QuarkV1alpha1().NetAppVolumeReplications(r.nvcClient.nvcNamespace).Delete(
			context.TODO(), r.nvr.Name, metav1.DeleteOptions{})
		if delNVRErr != nil {
			log.Errorf("deletion of VolumeReplication %v failed, err: %v", nvr.Name, delNVRErr)
		}
		return nil, err
	}
	return nvr, nil
}

func (r *nvrBackup) Delete() error {
	if !r.noBackupInProgress() {
		return fmt.Errorf("backup in progress, rejecting NetAppVolumeReplication delete request for %v", r.nvr.Name)
	}
	return r.nvcClient.quarkClient.QuarkV1alpha1().NetAppVolumeReplications(r.nvcClient.nvcNamespace).Delete(
		context.TODO(), r.nvr.Name, metav1.DeleteOptions{})
}

func (r *nvrBackup) Initialize(transferSnapshot string) error {
	// For backup, Create operation sets the spec.Mirrorstate to Mirrored so no need
	// for initialize operation
	return fmt.Errorf("unsupported operation for backup")
}

func (r *nvrBackup) Resync() error {
	return fmt.Errorf("unsupported operation for backup")
}

func (r *nvrBackup) Break() error {
	return fmt.Errorf("unsupported operation for backup")
}

// TransferSnapshot API will be used by SDE to start the transfer of subsequent snapshots (2nd snapshot
// onwards) after the first snapshot has already been transferred through the Create API.
// Backup uses spec/status.SnapshotToTransfer field
func (r *nvrBackup) TransferSnapshot(snapshotName string) error {
	if r.nvr.Spec.SnapshotToTransfer == snapshotName {
		return nil
	}
	if !NetAppVolumeReplicationMirrored(r.nvr) {
		return fmt.Errorf("volume replication state must be \"%v\". Current state is \"%v\"",
			quarkv1alpha1.MirrorStateMirrored, r.nvr.Status.MirrorState)
	}
	if !r.noBackupInProgress() {
		return fmt.Errorf("cannot initiate a new transfer while another transfer or abort is in progress")
	}
	volumeReplicationCopy := r.nvr.DeepCopy()
	volumeReplicationCopy.Spec.SnapshotToTransfer = snapshotName
	_, err := r.nvcClient.NetAppVolumeReplicationUpdate(volumeReplicationCopy)
	return err
}

func (r *nvrBackup) TransferAbort() error {
	if r.nvr.Spec.MirrorState == quarkv1alpha1.MirrorStateQuiesced {
		return nil
	}
	if r.noBackupInProgress() {
		return fmt.Errorf("no backup transfer currently in progress, nothing to abort")
	}
	r.nvr.Spec.AbortTransferCounter += 1
	r.nvr.Spec.MirrorState = quarkv1alpha1.MirrorStateQuiesced

	_, err := r.nvcClient.NetAppVolumeReplicationUpdate(r.nvr)
	return err
}

// ScheduleModify is not needed in initial release - global scheduler is handling it
func (r *nvrBackup) ScheduleModify(schedule quarkv1alpha1.NetAppVolumeReplicationScheduleType) error {
	return fmt.Errorf("unsupported operation for backup")
}

// - Cannot rely only on ReplicationStatus as there is no guarantee we will see the
// ReplicationStatusTransferring state since nvcclient essentially polls for this value and we
// might miss it.
// - Matching SnapshotToTransfer value is the only sureshot way of determining when backup is
// in progress along with ensuring ReplicationStatusIdle status.
// if r.nvr.Status.ReplicationStatus is nil it indicates a newly create nvr and no backup is in progress
func (r *nvrBackup) noBackupInProgress() bool {
	return r.nvr.Status.ReplicationStatus == "" ||
		((r.nvr.Spec.SnapshotToTransfer == r.nvr.Status.SnapshotToTransfer) &&
			(r.nvr.Status.ReplicationStatus == quarkv1alpha1.ReplicationStatusIdle))
}

func (r *nvrBackup) TransferResume(snapshotName string) error {
	// r.nvr.Spec.MirrorState is quarkv1alpha1.MirrorStateMirrored, resume already issued, no-op here
	if r.nvr.Spec.MirrorState == quarkv1alpha1.MirrorStateMirrored {
		return nil
	}
	if r.nvr.Spec.MirrorState != quarkv1alpha1.MirrorStateQuiesced {
		return fmt.Errorf("current mirrorState \"%v\" is not \"%v\", nothing to resume", r.nvr.Spec.MirrorState, quarkv1alpha1.MirrorStateQuiesced)
	}
	r.nvr.Spec.MirrorState = quarkv1alpha1.MirrorStateMirrored
	if snapshotName != "" {
		r.nvr.Spec.SnapshotToTransfer = snapshotName
	}
	_, err := r.nvcClient.NetAppVolumeReplicationUpdate(r.nvr)
	return err
}
