package v2

import (
	"context"
	"fmt"

	"k8s.io/apimachinery/pkg/api/errors"

	uuid "github.com/satori/go.uuid"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	quarkv1alpha1 "quark.netapp.io/volume-controller/pkg/apis/volumecontroller/v1alpha1"
	"quark.netapp.io/volume-controller/pkg/log"
	"quark.netapp.io/volume-controller/pkg/quarkconfig"
)

// nvrRestore implements the nvrInterface to support the restore usecase. Functionalities/APIs which are not
// planned for the first release of Integrated backup feature should return error until implemented.
type nvrRestore struct {
	nvrCommonImpl
}

func (r *nvrRestore) Create() (nvr *quarkv1alpha1.NetAppVolumeReplication, err error) {

	// This storage pool refers to a cluster-local storage pool
	storagePoolUUIDStr := r.nvr.Labels[quarkconfig.StoragePoolLabel]
	_, err = r.nvcClient.NetAppStoragePoolGet(storagePoolUUIDStr)
	if err != nil {
		return nil, fmt.Errorf("error getting StoragePool %v: %v", storagePoolUUIDStr, err)
	}

	if r.nvr.Spec.Destination == nil {
		return nil, fmt.Errorf("Destination field in spec in nil, cannot process create request")
	}
	// For restore, destination is of type NetAppVolume.
	// Only validate that it exists. We don't validate if this volume size is adequate to
	// fit the entire backup being restored since the original backup could exist on a
	// different cluster. If that is the case, we won't have access to associated
	// NetAppVolumeCloudBackupList CR which will have info (size) about this particular backup.
	// Another way to validate this would be to authenticate to the objectstore and
	// query endpointUUID information which will contain information about all the
	// available backups. As a design decision, we won't do this, and let dmap/SDE verify
	// volume size
	destVolUUIDStr := r.nvr.Spec.Destination.UUID
	destVol, err := r.nvcClient.NetAppVolumeGet(destVolUUIDStr)
	if err != nil {
		return nil, fmt.Errorf("error getting destination NetAppVolume %v: %v", destVolUUIDStr, err)
	}
	// Check if destination volume belongs to the storagepool supplied in the replication cr
	if destVol.Labels[quarkconfig.StoragePoolLabel] != storagePoolUUIDStr {
		return nil, fmt.Errorf("destination volume's storage pool (%v) not the same as storage pool (%v) supplied in the volume-replication resource", destVol.Labels[quarkconfig.StoragePoolLabel], storagePoolUUIDStr)
	}

	// Check if destination volume is data-protected
	if destVol.Status.DataProtection != quarkv1alpha1.DataProtectionEnabled {
		return nil, fmt.Errorf("unable to initiate restore to destination volume %v, as it is not data-protected", destVolUUIDStr)
	}

	// Validate if supplied objectstore is valid.
	// Since this is restore operation, the source backup could be from the same cluster
	// or a different cluster. If same cluster, the objectstore object will be pre-existing
	// since it was backed up to from the current cluster. If a different cluster, we need an
	// ObjectStore configuration associated with the bucket that has the snapshot to be restored.
	// SDE would have created this ObjectStore configuration on the current cluster before issuing
	// the restore request.
	objectStoreUUIDStr := r.nvr.Labels[quarkconfig.ObjectStoreLabel]
	objectStoreUUID, _ := uuid.FromString(objectStoreUUIDStr)
	_, err = r.nvcClient.NetAppObjectStoreGet(objectStoreUUID)
	if err != nil {
		return nil, fmt.Errorf("error getting ObjectStore %v: %v", objectStoreUUIDStr, err)
	}

	// Update dp attribute on volume
	// Note: dataProtection field on a volume is not reconciled as part of vol-reconcile loop. It is as part
	// of the volume health check routine. Snapmirror will set status.DP to r/w once restore is complete.
	// In the case where this NVR already exists, the volume's dp will already be disabled in which case
	// no action will be taken.
	origVolDPVal := destVol.Spec.DataProtection
	if destVol.Spec.DataProtection == quarkv1alpha1.DataProtectionEnabled {
		if volErr := r.nvcClient.netAppVolumeDataProtectionUpdate(
			r.nvr.Spec.Destination.UUID,
			quarkv1alpha1.DataProtectionDisabled); volErr != nil {
			log.WithFields(log.Fields{
				"destinationVolume": r.nvr.Spec.Destination.UUID,
				"volumeReplication": r.nvr.Name,
				"error":             volErr.Error(),
			}).Error("unable to disable dataProtection for volume")
			return nil, volErr
		}
	}

	// r.nvr.Spec.sourceUUID is the endpoint UUID, which is something we won't validate at this layer since it would
	// involve authenticating with the cloud bucket (ObjectStore) and querying contents.
	// We will pass sourceUUID as is and let dmap error out if supplied sourceUUID (endpoint UUID) is invalid.
	// dmap needs to authenticate with the objectstore anyway to read/write volume data.
	// Bad sourceUUID value will be communicated by dmap by a condition on the CR.
	nvr, err = r.nvcClient.quarkClient.QuarkV1alpha1().NetAppVolumeReplications(r.nvcClient.nvcNamespace).Create(
		context.TODO(), r.nvr, metav1.CreateOptions{})
	if err != nil {
		// Check for conflict error, in which case we don't want to reset vol's dp attribute
		if errors.IsAlreadyExists(err) {
			log.WithFields(log.Fields{
				"destinationVolume": r.nvr.Spec.Destination.UUID,
				"volumeReplication": r.nvr.Name,
			}).Warn("NetAppVolumeReplication (restore) CR already exists")
		} else {
			// Restore volume's dp attribute
			if volErr := r.nvcClient.netAppVolumeDataProtectionUpdate(
				r.nvr.Spec.Destination.UUID,
				origVolDPVal); volErr != nil {
				log.WithFields(log.Fields{
					"destinationVolume": r.nvr.Spec.Destination.UUID,
					"volumeReplication": r.nvr.Name,
					"error":             volErr.Error(),
					"dataProtection":    origVolDPVal,
				}).Error("unable to reset dataProtection value for volume")
			}
		}
		// Return all errors, including conflict error and let caller decide what to do with it
		return nil, err
	}
	return nvr, nil
}

func (r *nvrRestore) Delete() error {
	if !r.noRestoreInProgress() {
		return fmt.Errorf("restore in progress, rejecting NetAppVolumeReplication delete request for %v", r.nvr.Name)
	}
	return r.nvcClient.quarkClient.QuarkV1alpha1().NetAppVolumeReplications(r.nvcClient.nvcNamespace).Delete(
		context.TODO(), r.nvr.Name, metav1.DeleteOptions{})
}

func (r *nvrRestore) Initialize(transferSnapshot string) error {
	return fmt.Errorf("unsupported operation for restore")
}

func (r *nvrRestore) Resync() error {
	return fmt.Errorf("unsupported operation for restore")
}

func (r *nvrRestore) Break() error {
	return fmt.Errorf("unsupported operation for restore")
}

// TransferSnapshot API will be used by SDE to start the transfer of subsequent snapshots (2nd snapshot
// onwards) after the first snapshot has already been transferred through the Create API.
// Restore uses spec/status.SnapshotToTransfer field
func (r *nvrRestore) TransferSnapshot(snapshotName string) error {
	if r.nvr.Spec.SnapshotToTransfer == snapshotName {
		return nil
	}
	// status should be "Restored" since this is a subsequent snapshot transfer request
	if !NetAppVolumeReplicationRestored(r.nvr) {
		return fmt.Errorf("to initiate subsequent snapshot restore, volume replication state must be %v. Current state is %v",
			quarkv1alpha1.MirrorStateRestored, r.nvr.Status.MirrorState)
	}

	if !r.noRestoreInProgress() {
		return fmt.Errorf("cannot initiate a new transfer while another transfer or abort is in progress")
	}

	volumeReplicationCopy := r.nvr.DeepCopy()
	volumeReplicationCopy.Spec.SnapshotToTransfer = snapshotName
	_, err := r.nvcClient.NetAppVolumeReplicationUpdate(volumeReplicationCopy)
	return err
}

func (r *nvrRestore) TransferAbort() error {
	if r.nvr.Spec.MirrorState == quarkv1alpha1.MirrorStateQuiesced {
		return nil
	}
	if r.noRestoreInProgress() {
		fmt.Errorf("no restore operations in progress, nothing to abort")
	}
	r.nvr.Spec.AbortTransferCounter += 1
	r.nvr.Spec.MirrorState = quarkv1alpha1.MirrorStateQuiesced

	_, err := r.nvcClient.NetAppVolumeReplicationUpdate(r.nvr)
	return err
}

func (r *nvrRestore) ScheduleModify(schedule quarkv1alpha1.NetAppVolumeReplicationScheduleType) error {
	return fmt.Errorf("unsupported operation for restore")
}

// It is more straightforward to determine that no restore is in progress than to figure out
// if one is in progress or if the restore operation has an error condition on it.
// - Cannot use MirrorState as status.MirrorState will always be "Restored" after the first
// transfer
// - Cannot rely only on ReplicationStatus as there is no guarantee we will see the
// ReplicationStatusTransferring state since nvcclient essentially polls for this value and we
// might miss it.
// - Matching SnapshotToTransfer value is the only sureshot way of determining when restore is
// in progress along with ensuring ReplicationStatusIdle status.
// if r.nvr.Status.ReplicationStatus is nil it indicates a newly create nvr and no restore is in progress
func (r *nvrRestore) noRestoreInProgress() bool {
	return r.nvr.Status.ReplicationStatus == "" ||
		((r.nvr.Spec.SnapshotToTransfer == r.nvr.Status.SnapshotToTransfer) &&
			(r.nvr.Status.ReplicationStatus == quarkv1alpha1.ReplicationStatusIdle))
}

func (r *nvrRestore) TransferResume(snapshotName string) error {
	// r.nvr.Spec.MirrorState is quarkv1alpha1.MirrorStateMirrored, resume already issued, no-op here
	if r.nvr.Spec.MirrorState == quarkv1alpha1.MirrorStateRestored {
		return nil
	}
	if r.nvr.Spec.MirrorState != quarkv1alpha1.MirrorStateQuiesced {
		return fmt.Errorf("current mirrorState \"%v\" is not \"%v\", nothing to resume", r.nvr.Spec.MirrorState, quarkv1alpha1.MirrorStateQuiesced)
	}
	r.nvr.Spec.MirrorState = quarkv1alpha1.MirrorStateRestored
	_, err := r.nvcClient.NetAppVolumeReplicationUpdate(r.nvr)
	return err
}
